"""Collection of stastical methods: Transformation and Skill Scores."""

from scipy import stats

__author__ = 'Lukas Leufen, Felix Kleinert'
__date__ = '2019-10-23'

import numpy as np
import xarray as xr
import pandas as pd
from typing import Union, Tuple, Dict, List
import itertools

Data = Union[xr.DataArray, pd.DataFrame]


def apply_inverse_transformation(data: Data, method: str = "standardise", mean: Data = None, std: Data = None,
                                 max: Data = None, min: Data = None, feature_range: Data = None) -> Data:
    """
    Apply inverse transformation for given statistics.

    :param data: transform this data back
    :param method: transformation method (optional)
    :param mean: mean of transformation (optional)
    :param std: standard deviation of transformation (optional)
    :param max: maximum value for min/max transformation (optional)
    :param min: minimum value for min/max transformation (optional)

    :return: inverse transformed data
    """
    if method == 'standardise':  # pragma: no branch
        return standardise_inverse(data, mean, std)
    elif method == 'centre':  # pragma: no branch
        return centre_inverse(data, mean)
    elif method == 'min_max':  # pragma: no branch
        return min_max_inverse(data, min, max, feature_range)
    elif method == "log":
        return log_inverse(data, mean, std)
    else:
        raise NotImplementedError


def standardise(data: Data, dim: Union[str, int]) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Standardise a xarray.dataarray (along dim) or pandas.DataFrame (along axis) with mean=0 and std=1.

    :param data: data to standardise
    :param dim: name (xarray) or axis (pandas) of dimension which should be standardised
    :return: standardised data, and dictionary with keys method, mean, and standard deviation
    """
    return (data - data.mean(dim)) / data.std(dim), {"mean": data.mean(dim), "std": data.std(dim),
                                                     "method": "standardise"}


def standardise_inverse(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply inverse function of `standardise` on data and therefore vanishes the standardising.

    :param data: standardised data
    :param mean: mean of standardisation
    :param std: standard deviation of transformation

    :return: inverse standardised data
    """
    return data * std + mean


def standardise_apply(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply `standardise` on data using given mean and std.

    :param data: data to transform
    :param mean: mean to use for transformation
    :param std: standard deviation for transformation

    :return: transformed data
    """
    return (data - mean) / std


def centre(data: Data, dim: Union[str, int]) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Centre a xarray.dataarray (along dim) or pandas.DataFrame (along axis) to mean=0.

    :param data: data to centre
    :param dim: name (xarray) or axis (pandas) of dimension which should be centred

    :return: centred data, and dictionary with keys method, and mean
    """
    return data - data.mean(dim), {"mean": data.mean(dim), "method": "centre"}


def centre_inverse(data: Data, mean: Data) -> Data:
    """
    Apply inverse function of `centre` and therefore add given values of mean to data.

    :param data: data to apply inverse centering
    :param mean: mean to use for inverse transformation

    :return: inverted centering transformation data
    """
    return data + mean


def centre_apply(data: Data, mean: Data) -> Data:
    """
    Apply `centre` on data using given mean.

    :param data: data to transform
    :param mean: mean to use for transformation

    :return: transformed data
    """
    return data - mean


def min_max(data: Data, dim: Union[str, int], feature_range: Tuple = (0, 1)) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Apply min/max scaling using (x - x_min) / (x_max - x_min). Returned data is in interval [0, 1].

    :param data: data to transform
    :param dim: name (xarray) or axis (pandas) of dimension which should be centred
    :param feature_range: scale data to any interval given in feature range. Default is scaling on interval [0, 1].
    :return: transformed data, and dictionary with keys method, min, and max
    """
    d_max = data.max(dim)
    d_min = data.min(dim)
    d_scaled = (data - d_min) / (d_max - d_min) * (max(feature_range) - min(feature_range)) + min(feature_range)
    return d_scaled, {"min": d_min, "max": d_max, "method": "min_max", "feature_range": feature_range}


def min_max_inverse(data: Data, _min: Data, _max: Data, feature_range: Tuple = (0, 1)) -> Data:
    """
    Apply inverse transformation of `min_max` scaling.

    :param data: data to apply inverse scaling
    :param _min: minimum value to use for min/max scaling
    :param _max: maximum value to use for min/max scaling
    :param feature_range: scale data to any interval given in feature range. Default is scaling on interval [0, 1].
    :return: inverted min/max scaled data
    """
    return (data - min(feature_range)) / (max(feature_range) - min(feature_range)) * (_max - _min) + _min


def min_max_apply(data: Data, _min: Data, _max: Data, feature_range: Data = (0, 1)) -> Data:
    """
    Apply `min_max` scaling with given minimum and maximum.

    :param data: data to apply scaling
    :param _min: minimum value to use for min/max scaling
    :param _max: maximum value to use for min/max scaling
    :param feature_range: scale data to any interval given in feature range. Default is scaling on interval [0, 1].
    :return: min/max scaled data
    """
    if not isinstance(feature_range, xr.DataArray):
        return (data - _min) / (_max - _min) * (max(feature_range) - min(feature_range)) + min(feature_range)
    else:
        return (data - _min) / (_max - _min) * (feature_range.max() - feature_range.min()) + feature_range.min()


def log(data: Data, dim: Union[str, int]) -> Tuple[Data, Dict[(str, Data)]]:
    """
    Apply logarithmic transformation (and standarization) to data. This method first uses the logarithm for
    transformation and second applies the `standardise` method additionally. A logarithmic function numpy's log1p is
    used (`res = log(1+x)`) instead of the pure logarithm to be applicable to values of 0 too.

    :param data: transform this data
    :param dim: name (xarray) or axis (pandas) of dimension which should be transformed
    :return: transformed data, and option dictionary with keys method, mean, and std
    """
    transformed_standardized, opts = standardise(np.log1p(data), dim)
    opts.update({"method": "log"})
    return transformed_standardized, opts


def log_inverse(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply inverse log transformation (therefore exponential transformation). Because `log` is using `np.log1p` this
    method is based on the equivalent method `np.exp1m`. Data are first rescaled using `standardise_inverse` and then
    given to the exponential function.

    :param data: apply inverse log transformation on this data
    :param mean: mean of the standarization
    :param std: std of the standarization
    :return: inverted data
    """
    data_rescaled = standardise_inverse(data, mean, std)
    return np.expm1(data_rescaled)


def log_apply(data: Data, mean: Data, std: Data) -> Data:
    """
    Apply numpy's log1p on given data. Further information can be found in description of `log` method.

    :param data: transform this data
    :param mean: mean of the standarization
    :param std: std of the standarization
    :return: transformed data
    """
    return standardise_apply(np.log1p(data), mean, std)


def mean_squared_error(a, b, dim=None):
    """Calculate mean squared error."""
    return np.square(a - b).mean(dim)


def mean_absolute_error(a, b, dim=None):
    """Calculate mean absolute error."""
    return np.abs(a - b).mean(dim)


def calculate_error_metrics(a, b, dim):
    """Calculate MSE, RMSE, and MAE. Additionally return number of used values for calculation."""
    mse = mean_squared_error(a, b, dim)
    rmse = np.sqrt(mse)
    mae = mean_absolute_error(a, b, dim)
    n = (a - b).notnull().sum(dim)
    return {"mse": mse, "rmse": rmse, "mae": mae, "n": n}


class SkillScores:
    r"""
    Calculate different kinds of skill scores.

    Skill score on MSE:
        Calculate skill score based on MSE for given forecast, reference and observations.

        .. math::

            \text{SkillScore} = 1 - \frac{\text{MSE(obs, for)}}{\text{MSE(obs, ref)}}

        To run:

        .. code-block:: python

            skill_scores = SkillScores(None).general_skill_score(data, observation_name, forecast_name, reference_name)

    Competitive skill score:
        Calculate skill scores to highlight differences between forecasts. This skill score is also based on the MSE.
        Currently required forecasts are CNN, OLS and persi, as well as the observation obs.

        .. code-block:: python

            skill_scores_class = SkillScores(internal_data)  # must contain columns CNN, OLS, persi and obs.
            skill_scores = skill_scores_class.skill_scores(window_lead_time=3)

    Skill score according to Murphy:
        Follow climatological skill score definition of Murphy (1988). External data is data from another time period
        than the internal data set on initialisation. In other terms, this should be the train and validation data
        whereas the external data is the test data. This sounds perhaps counter-intuitive, but if a skill score is
        evaluated to a model to another, this must be performed on test data set. Therefore, for this case the foreign
        data is test.

        .. code-block:: python

            skill_scores_class = SkillScores(external_data)  # must contain columns obs and CNN.
            skill_scores_clim = skill_scores_class.climatological_skill_scores(internal_data, window_lead_time=3)

    """
    models_default = ["cnn", "persi", "ols"]

    def __init__(self, external_data: Union[Data, None], models=None, observation_name="obs", ahead_dim="ahead"):
        """Set internal data."""
        self.external_data = external_data
        self.models = self.set_model_names(models)
        self.observation_name = observation_name
        self.ahead_dim = ahead_dim

    def set_model_names(self, models: List[str]) -> List[str]:
        """Either use given models or use defaults."""
        if models is None:
            models = self.models_default
        return self._reorder(models)

    @staticmethod
    def _reorder(model_list: List[str]) -> List[str]:
        """Set elements persi and obs at the very end of given list."""
        for element in ["persi", "obs"]:
            if element in model_list:
                model_list.append(model_list.pop(model_list.index(element)))
        return model_list

    def get_model_name_combinations(self):
        """Return all combinations of two models as tuple and string."""
        combinations = list(itertools.combinations(self.models, 2))
        combination_strings = [f"{first} - {second}" for (first, second) in combinations]
        return combinations, combination_strings

    def skill_scores(self) -> [pd.DataFrame, pd.DataFrame]:
        """
        Calculate skill scores for all combinations of model names.

        :return: skill score for each comparison and forecast step
        """
        ahead_names = list(self.external_data[self.ahead_dim].data)
        combinations, combination_strings = self.get_model_name_combinations()
        skill_score = pd.DataFrame(index=combination_strings)
        count = pd.DataFrame(index=combination_strings)
        for iahead in ahead_names:
            data = self.external_data.sel({self.ahead_dim: iahead})
            skill_score[iahead] = [self.general_skill_score(data,
                                                            forecast_name=first,
                                                            reference_name=second,
                                                            observation_name=self.observation_name)
                                   for (first, second) in combinations]
            count[iahead] = [self.get_count(data,
                                            forecast_name=first,
                                            reference_name=second,
                                            observation_name=self.observation_name)
                             for (first, second) in combinations]
        return skill_score, count

    def climatological_skill_scores(self, internal_data: Data, forecast_name: str) -> xr.DataArray:
        """
        Calculate climatological skill scores according to Murphy (1988).

        Calculate all CASES I - IV and terms [ABC][I-IV]. Internal data has to be set by initialisation, external data
        is part of parameters.

        :param internal_data: internal data
        :param forecast_name: name of the forecast to use for this calculation (must be available in `data`)

        :return: all CASES as well as all terms
        """
        if self.external_data is not None:
            ahead_names = list(self.external_data[self.ahead_dim].data)
        else:
            ahead_names = list(internal_data[self.ahead_dim].data)

        all_terms = ['AI', 'AII', 'AIII', 'AIV', 'BI', 'BII', 'BIV', 'CI', 'CIV', 'CASE I', 'CASE II', 'CASE III',
                     'CASE IV']
        skill_score = xr.DataArray(np.full((len(all_terms), len(ahead_names)), np.nan), coords=[all_terms, ahead_names],
                                   dims=['terms', self.ahead_dim])

        for iahead in ahead_names:
            data = internal_data.sel({self.ahead_dim: iahead})

            skill_score.loc[["CASE I", "AI", "BI", "CI"], iahead] = np.stack(self._climatological_skill_score(
                data, mu_type=1, forecast_name=forecast_name, observation_name=self.observation_name).values.flatten())

            skill_score.loc[["CASE II", "AII", "BII"], iahead] = np.stack(self._climatological_skill_score(
                data, mu_type=2, forecast_name=forecast_name, observation_name=self.observation_name).values.flatten())

            if self.external_data is not None and self.observation_name in self.external_data.coords["type"]:
                external_data = self.external_data.sel({self.ahead_dim: iahead, "type": [self.observation_name]})
                skill_score.loc[["CASE III", "AIII"], iahead] = np.stack(self._climatological_skill_score(
                    data, mu_type=3, forecast_name=forecast_name, observation_name=self.observation_name,
                    external_data=external_data).values.flatten())

                skill_score.loc[["CASE IV", "AIV", "BIV", "CIV"], iahead] = np.stack(self._climatological_skill_score(
                    data, mu_type=4, forecast_name=forecast_name, observation_name=self.observation_name,
                    external_data=external_data).values.flatten())

        return skill_score

    def _climatological_skill_score(self, internal_data, observation_name, forecast_name, mu_type=1,
                                    external_data=None):
        kwargs = {"external_data": external_data} if external_data is not None else {}
        return self.__getattribute__(f"skill_score_mu_case_{mu_type}")(internal_data, observation_name, forecast_name,
                                                                       **kwargs)

    def general_skill_score(self, data: Data, forecast_name: str, reference_name: str,
                            observation_name: str = None, dim: str = "index") -> np.ndarray:
        r"""
        Calculate general skill score based on mean squared error.

        :param data: internal data containing data for observation, forecast and reference
        :param observation_name: name of observation
        :param forecast_name: name of forecast
        :param reference_name: name of reference

        :return: skill score of forecast
        """
        if observation_name is None:
            observation_name = self.observation_name
        data = data.dropna(dim)
        observation = data.sel(type=observation_name)
        forecast = data.sel(type=forecast_name)
        reference = data.sel(type=reference_name)
        mse = mean_squared_error
        skill_score = 1 - mse(observation, forecast, dim=dim) / mse(observation, reference, dim=dim)
        return skill_score.values

    def get_count(self, data: Data, forecast_name: str, reference_name: str,
                            observation_name: str = None) -> np.ndarray:
        r"""
        Calculate general skill score based on mean squared error.

        :param data: internal data containing data for observation, forecast and reference
        :param observation_name: name of observation
        :param forecast_name: name of forecast
        :param reference_name: name of reference

        :return: skill score of forecast
        """
        data = data.dropna("index")
        return data.count("index").max().values

    def skill_score_pre_calculations(self, data: Data, observation_name: str, forecast_name: str) -> Tuple[np.ndarray,
                                                                                                           np.ndarray,
                                                                                                           np.ndarray,
                                                                                                           Data,
                                                                                                           Dict[
                                                                                                               str, Data]]:
        """
        Calculate terms AI, BI, and CI, mean, variance and pearson's correlation and clean up data.

        The additional information on mean, variance and pearson's correlation (and the p-value) are returned as
        dictionary with the corresponding keys mean, sigma, r and p.

        :param data: internal data to use for calculations
        :param observation_name: name of observation
        :param forecast_name: name of forecast

        :returns: Terms AI, BI, and CI, internal data without nans and mean, variance, correlation and its p-value
        """
        data = data.sel(type=[observation_name, forecast_name]).drop(self.ahead_dim)
        data = data.dropna("index")

        mean = data.mean("index")
        sigma = np.sqrt(data.var("index"))
        r, p = stats.pearsonr(*[data.sel(type=n).values.flatten() for n in [forecast_name, observation_name]])

        AI = np.array(r ** 2)
        BI = ((r - (sigma.sel(type=forecast_name, drop=True) / sigma.sel(type=observation_name,
                                                                         drop=True))) ** 2).values
        CI = (((mean.sel(type=forecast_name, drop=True) - mean.sel(type=observation_name, drop=True)) / sigma.sel(
            type=observation_name, drop=True)) ** 2).values

        suffix = {"mean": mean, "sigma": sigma, "r": r, "p": p}
        return AI, BI, CI, data, suffix

    def skill_score_mu_case_1(self, internal_data, observation_name, forecast_name):
        """Calculate CASE I."""
        AI, BI, CI, data, _ = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        skill_score = np.array(AI - BI - CI)
        return pd.DataFrame({"skill_score": [skill_score], "AI": [AI], "BI": [BI], "CI": [CI]}).to_xarray().to_array()

    def skill_score_mu_case_2(self, internal_data, observation_name, forecast_name):
        """Calculate CASE II."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        monthly_mean = self.create_monthly_mean_from_daily_data(data)
        data = xr.concat([data, monthly_mean], dim="type")
        sigma = suffix["sigma"]
        sigma_monthly = np.sqrt(monthly_mean.var())
        r, p = stats.pearsonr(*[data.sel(type=n).values.flatten() for n in [observation_name, observation_name + "X"]])
        AII = np.array(r ** 2)
        BII = ((r - sigma_monthly / sigma.sel(type=observation_name, drop=True)) ** 2).values
        skill_score = np.array((AI - BI - CI - AII + BII) / (1 - AII + BII))
        return pd.DataFrame({"skill_score": [skill_score], "AII": [AII], "BII": [BII]}).to_xarray().to_array()

    def skill_score_mu_case_3(self, internal_data, observation_name, forecast_name, external_data=None):
        """Calculate CASE III."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        mean, sigma = suffix["mean"], suffix["sigma"]
        AIII = (((external_data.mean().values - mean.sel(type=observation_name, drop=True)) / sigma.sel(
            type=observation_name, drop=True)) ** 2).values
        skill_score = np.array((AI - BI - CI + AIII) / (1 + AIII))
        return pd.DataFrame({"skill_score": [skill_score], "AIII": [AIII]}).to_xarray().to_array()

    def skill_score_mu_case_4(self, internal_data, observation_name, forecast_name, external_data=None):
        """Calculate CASE IV."""
        AI, BI, CI, data, suffix = self.skill_score_pre_calculations(internal_data, observation_name, forecast_name)
        monthly_mean_external = self.create_monthly_mean_from_daily_data(external_data, index=data.index)
        data = xr.concat([data, monthly_mean_external], dim="type").dropna(dim="index")
        mean, sigma = suffix["mean"], suffix["sigma"]
        mean_external = monthly_mean_external.mean()
        sigma_external = np.sqrt(monthly_mean_external.var())
        r_mu, p_mu = stats.pearsonr(
            *[data.sel(type=n).values.flatten() for n in [observation_name, observation_name + "X"]])
        AIV = np.array(r_mu ** 2)
        BIV = ((r_mu - sigma_external / sigma.sel(type=observation_name, drop=True)) ** 2).values
        CIV = (((mean_external - mean.sel(type=observation_name, drop=True)) / sigma.sel(type=observation_name,
                                                                                         drop=True)) ** 2).values
        skill_score = np.array((AI - BI - CI - AIV + BIV + CIV) / (1 - AIV + BIV + CIV))
        return pd.DataFrame(
            {"skill_score": [skill_score], "AIV": [AIV], "BIV": [BIV], "CIV": CIV}).to_xarray().to_array()

    @staticmethod
    def create_monthly_mean_from_daily_data(data, columns=None, index=None):
        """
        Calculate average for each month and save as daily values with flag 'X'.

        :param data: data to average
        :param columns: columns to work on (all columns from given data are used if empty)
        :param index: index of returned data (index of given data is used if empty)

        :return: data containing monthly means in daily resolution
        """
        if columns is None:
            columns = data.type.values
        if index is None:
            index = data.index
        coordinates = [index, [v + "X" for v in list(columns)]]
        empty_data = np.full((len(index), len(columns)), np.nan)
        monthly_mean = xr.DataArray(empty_data, coords=coordinates, dims=["index", "type"])
        mu = data.groupby("index.month").mean()

        for month in mu.month:
            monthly_mean[monthly_mean.index.dt.month == month, :] = mu[mu.month == month].values.flatten()

        return monthly_mean


def create_single_bootstrap_realization(data: xr.DataArray, dim_name_time: str) -> xr.DataArray:
    """
    Return a bootstraped realization of data
    :param data: data from which to draw ONE bootstrap realization
    :param dim_name_time: name of time dimension
    :return: bootstrapped realization of data
    """

    num_of_blocks = data.coords[dim_name_time].shape[0]
    boot_idx = np.random.choice(num_of_blocks, size=num_of_blocks, replace=True)
    return data.isel({dim_name_time: boot_idx})


def calculate_average(data: xr.DataArray, **kwargs) -> xr.DataArray:
    """
    Calculate mean of data
    :param data: data for which to calculate mean
    :return: mean of data
    """
    return data.mean(**kwargs)


def create_n_bootstrap_realizations(data: xr.DataArray, dim_name_time: str, dim_name_model: str, n_boots: int = 1000,
                                    dim_name_boots: str = 'boots') -> xr.DataArray:
    """
    Create n bootstrap realizations and calculate averages across realizations

    :param data: original data from which to create bootstrap realizations
    :param dim_name_time: name of time dimension
    :param dim_name_model: name of model dimension
    :param n_boots: number of bootstap realizations
    :param dim_name_boots: name of bootstap dimension
    :return:
    """
    res_dims = [dim_name_boots]
    dims = list(data.dims)
    coords = {dim_name_boots: range(n_boots), dim_name_model: data.coords[dim_name_model] }
    if len(dims) > 1:
        res_dims = res_dims + dims[1:]
    res = xr.DataArray(np.nan, dims=res_dims, coords=coords)
    for boot in range(n_boots):
        res[boot] = (calculate_average(
            create_single_bootstrap_realization(data, dim_name_time=dim_name_time),
            dim=dim_name_time, skipna=True))
    return res









